#!/usr/bin/env python3
# PYTHON_ARGCOMPLETE_OK
# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

from __future__ import annotations

import argparse
import json
import os
import shutil
import sys
from collections.abc import Collection, Iterable, Iterator, MutableSequence, Sequence
from contextlib import contextmanager
from functools import cache
from pathlib import Path
from tempfile import TemporaryDirectory
from textwrap import dedent
from typing import IO, Any, cast

import license_expression
from zstarfile import ZSTarfile

from go_vendor_tools import __version__
from go_vendor_tools.archive import get_toplevel_directory
from go_vendor_tools.config.base import load_config
from go_vendor_tools.config.licenses import (
    LicenseConfig,
    LicenseEntry,
    create_license_config,
)
from go_vendor_tools.gomod import get_unlicensed_mods
from go_vendor_tools.hashing import get_hash
from go_vendor_tools.license_detection.base import LicenseData, LicenseDetector
from go_vendor_tools.license_detection.load import DETECTORS, get_detctors
from go_vendor_tools.licensing import compare_licenses, simplify_license
from go_vendor_tools.specfile_sources import get_path_and_output_from_specfile

try:
    import tomlkit
except ImportError:
    HAS_TOMLKIT = False
else:
    HAS_TOMLKIT = True
    from go_vendor_tools.cli.utils import load_tomlkit_if_exists
try:
    import argcomplete
except ImportError:
    HAS_ARGCOMPLETE = False
else:
    HAS_ARGCOMPLETE = True

COLOR: bool | None = None
RED = "\033[31m"  # ]
CLEAR = "\033[0m"  # ]


def red(__msg: str, /, *, file: IO[str] = sys.stdout) -> None:
    color = COLOR
    if color is None:
        color = file.isatty()
    print(f"{RED if color else ''}{__msg}{CLEAR if color else ''}", file=file)


def split_kv_options(kv_config: list[str]) -> dict[str, str]:
    results: dict[str, str] = {}
    for opt in kv_config:
        if ";" in opt:
            results |= split_kv_options(opt.split(";"))
        else:
            key, _, value = opt.partition("=")
            results[key] = value
    return results


def choose_license_detector(
    choice: str | None, license_config: LicenseConfig, kv_config: list[str] | None
) -> LicenseDetector:
    kv_config = kv_config or []
    cli_config = split_kv_options(kv_config)
    available, missing = get_detctors(cli_config, license_config)
    if choice:
        if choice in missing:
            sys.exit(f"Failed to get detector {choice!r}: {missing[choice]}")
        return available[choice]
    if not available:
        print("Failed to load license detectors:", file=sys.stderr)
        for detector, err in missing.items():
            print(f"! {detector}: {err}")
        sys.exit()
    return next(iter(available.values()))


def _add_json_argument(parser: argparse.ArgumentParser, **kwargs) -> None:
    our_kwargs: dict[str, Any] = {
        "type": Path,
        "help": "Write license data to a JSON file",
    }
    kwargs = our_kwargs | kwargs
    parser.add_argument("--write-json", **kwargs)


def parseargs(argv: list[str] | None = None) -> argparse.Namespace:
    """
    Parse arguments and return an `argparse.Namespace`
    """
    parser = argparse.ArgumentParser(
        description="Handle licenses for vendored go projects"
    )
    parser.add_argument("--version", action="version", version=__version__)
    parser.add_argument("-c", "--config", type=Path, dest="config_path")
    parser.add_argument(
        "-C",
        "--path",
        "--directory",
        type=Path,
        dest="directory",
        action="append",
        help=dedent(
            """
        Can be one of the following:

        1. Top-level directory with a go.mod file and vendor directory
        2. If --use-archive is specified, treat as a tarball and unpack it. Can
           be specified multiple times to unpack multiple tarballs on
           top of one another.
        3. Path to a specfile.
           The paths to Source0 and Source1 will be automatically unpacked.
        """
        ),
    )
    parser.add_argument("--use-archive", action="store_true")
    parser.add_argument(
        "--color",
        action=argparse.BooleanOptionalAction,
        default=False if os.environ.get("NO_COLOR") else None,
    )
    parser.add_argument(
        "-d",
        "--detector",
        choices=DETECTORS,
        default=None,
        help="Choose a license detector. Choices: %(choices)s. Default: autodetect",
        dest="detector_name",
    )
    parser.add_argument(
        "-D",
        "--dc",
        "--detector-config",
        help="KEY=VALUE pairs to pass to the license detector."
        " Can be passed multiple times",
        dest="detector_config",
        action="append",
    )
    subparsers = parser.add_subparsers(dest="subcommand")
    subparsers.required = True
    report_parser = subparsers.add_parser("report", help="Main subcommand")
    report_parser.add_argument(
        "-i",
        "--ignore-undetected",
        action="store_true",
        help="Whether to show undetected licenses in the output",
    )
    report_parser.add_argument(
        "-L",
        "--ignore-unlicensed-mods",
        action="store_true",
        help="Whether to show Go modules without licenses in the output",
    )
    report_parser.add_argument(
        "--verify",
        help="Verify license expression to make sure it matches caluclated expression",
        metavar="EXPRESSION",
    )
    report_parser.add_argument(
        "--prompt",
        action=argparse.BooleanOptionalAction,
        help="Whether to prompt to fill in undetected licenses."
        " Default: %(default)s",
    )
    report_parser.add_argument(
        "mode",
        nargs="?",
        type=str,
        choices=("all", "expression", "list"),
        default="all",
    )
    _add_json_argument(report_parser)
    report_parser.add_argument(
        "--write-config", help="Write a base config.", action="store_true"
    )
    explict_parser = subparsers.add_parser(
        "explicit", help="Add manual license entry to a config file"
    )
    explict_parser.add_argument(
        "-f", "--file", dest="license_file", required=True, type=Path
    )
    explict_parser.add_argument("license_expression")
    install_parser = subparsers.add_parser(
        "install", description=f"INTERNAL: {install_command.__doc__}"
    )
    install_parser.add_argument(
        "--install-directory", dest="install_directory", type=Path, required=True
    )
    install_parser.add_argument(
        "--destdir", dest="install_destdir", type=Path, default=Path("/")
    )
    install_parser.add_argument(
        "--filelist", dest="install_filelist", type=Path, required=True
    )
    # TODO(gotmax23): Should we support writing JSON from the install command
    # or just reading it? _add_json_argument(install_parser)
    generate_buildrequires_parser = subparsers.add_parser(  # noqa F841
        "generate_buildrequires"
    )

    if HAS_ARGCOMPLETE:
        argcomplete.autocomplete(parser)
    args = parser.parse_args(argv)
    args.directory = list(map(Path, args.directory or (".")))
    if args.subcommand not in ("explicit",):
        loaded = load_config(
            args.config_path, allow_missing=getattr(args, "write_config", False)
        )
        args.config = loaded["licensing"]
        if not args.detector_name:
            args.detector_name = args.config["detector"]
    if args.subcommand in ("report", "install"):
        args.detector = choose_license_detector(
            args.detector_name, args.config, args.detector_config
        )
        # TODO(anyone): Replace the print if/when we implement more granular logging
        print("Using detector:", args.detector.NAME, file=sys.stderr)
    global COLOR  # noqa: PLW0603
    COLOR = args.color
    return args


def bullet_iterator(it: Iterable[object], bullet: str = "- ") -> Iterator[str]:
    for item in it:
        yield bullet + str(item)


def red_if_true(items: Collection[object], message: str, bullet: str = "- ") -> None:
    if not items:
        return
    print(message)
    red("\n".join(bullet_iterator(items, bullet)))


def paths_relative_to_list(paths: Collection[Path], directory: Path) -> list[Path]:
    return [path.resolve().relative_to(directory.resolve()) for path in paths]


def print_licenses(
    results: LicenseData,
    unlicensed_mods: Collection[Path],
    mode: str,
    show_undetected: bool,
    show_unlicensed,
    directory: Path,
) -> None:
    if mode in ("all", "list"):
        for (
            license_path,
            license_name,
        ) in results.license_map.items():
            print(f"{license_path}: {license_name}")
    if (
        results.undetected_licenses
        or unlicensed_mods
        or results.unmatched_extra_licenses
    ):
        if mode != "expression":
            print()
        if show_undetected:
            red_if_true(
                paths_relative_to_list(results.undetected_licenses, directory),
                "The following license files were found "
                "but the correct license identifier couldn't be determined:",
            )
        if show_unlicensed:
            red_if_true(
                paths_relative_to_list(unlicensed_mods, directory),
                "The following modules are missing license files:",
            )
        red_if_true(
            results.unmatched_extra_licenses,
            "The following license files that were specified in the configuration"
            " have changed:",
        )
    if mode == "list":
        return
    if mode != "expression":
        print()
    print(results.license_expression)


def write_license_json(data: LicenseData, file: Path) -> None:
    with file.open("w", encoding="utf-8") as fp:
        json.dump(data.to_jsonable(), fp)


@cache
def need_tomlkit(action="this action"):
    if not HAS_TOMLKIT:
        message = f"tomlkit is required for {action}. Please install it!"
        sys.exit(message)


def tomlkit_dump(obj: Any, path: Path) -> None:
    need_tomlkit()
    with path.open("w") as fp:
        tomlkit.dump(obj, fp)


def prompt_missing_licenses(
    data: LicenseData,
    entries: MutableSequence[LicenseEntry],
) -> tuple[LicenseData, MutableSequence[LicenseEntry]]:
    if not data.undetected_licenses:
        return data, entries
    print("Undetected licenses found! Please enter them manually.")
    undetected_licenses = set(data.undetected_licenses)
    license_map: dict[Path, str] = dict(data.license_map)
    for undetected in sorted(data.undetected_licenses):
        print(f"* Undetected license: {data.directory / undetected}")
        expression_str = input("Enter SPDX expression (or IGNORE): ")
        if expression_str == "IGNORE":
            undetected_licenses.remove(undetected)
            print("Ignoring...")
            continue
        expression: str = (
            str(simplify_license(expression_str)) if expression_str else ""
        )
        print(f"Expression simplified to {expression!r}")
        license_map[undetected] = expression
        entry_dict = LicenseEntry(
            path=str(undetected),
            sha256sum=get_hash(data.directory / undetected),
            expression=expression,
        )
        replace_entry(entries, entry_dict, undetected)
        undetected_licenses.remove(undetected)
    assert not undetected_licenses
    return (
        data.replace(undetected_licenses=undetected_licenses, license_map=license_map),
        entries,
    )


def _write_config_verify_path(config_path: Path | None) -> Path:
    if config_path:
        return config_path
    need_tomlkit("--write-config")

    default = Path.cwd() / "go-vendor-tools.toml"
    if default.is_file():
        sys.exit("--write-config: Please pass --config to write configuration file!")
    else:
        print(
            "WARNING --write-config: No --config path specified"
            f" Will write to {default}",
            file=sys.stderr,
        )
    return config_path or default


def get_report_write_config_data(
    config_path: Path | None, detector: LicenseDetector
) -> tuple[Path, tomlkit.TOMLDocument]:
    need_tomlkit("--write-config")
    new_config_path = _write_config_verify_path(config_path)
    loaded = load_tomlkit_if_exists(config_path)
    write_config_data = loaded.setdefault("licensing", {})
    write_config_data |= {"detector": detector.NAME}
    return new_config_path, loaded


def write_and_prompt_report_licenses(
    license_data: LicenseData, write_config_data: tomlkit.TOMLDocument
) -> LicenseData:
    # fmt: off
    license_config_list = (
        write_config_data
        .setdefault("licensing", {})
        .setdefault("licenses", tomlkit.aot() if HAS_TOMLKIT else [])
    )
    # fmt: on
    license_data, _ = prompt_missing_licenses(license_data, license_config_list)
    return license_data


@contextmanager
def handle_alternative_sources(
    directories: Sequence[Path], is_archive: bool
) -> Iterator[Path]:
    directories = list(directories)
    # The CLI code already checks that the value is greater than zero, so
    # assert is fine here
    assert directories
    if not is_archive and len(directories) > 1:
        sys.exit("Too many paths were passed!")

    if directories[0].suffix == ".spec":
        directories[:] = get_path_and_output_from_specfile(directories[0])
        is_archive = True
    if is_archive:
        with TemporaryDirectory() as _tmp:
            tmp = Path(_tmp)
            # Extract the first archive
            with ZSTarfile.open(directories[0]) as tar:
                first_toplevel = get_toplevel_directory(tar)
                if not first_toplevel:
                    sys.exit(f"{directories[0]} does not have a top-level directory!")
                print(f"Extracting {directories[0]}", file=sys.stderr)
                tar.extractall(tmp)
            for directory in directories[1:]:
                with ZSTarfile.open(directory) as tar:
                    toplevel = get_toplevel_directory(tar)
                    print(f"Extracting {directory}", file=sys.stderr)
                    tar.extractall(tmp if toplevel else tmp / first_toplevel)
            yield tmp / first_toplevel

    else:
        yield directories[0]


def report_command(args: argparse.Namespace) -> None:
    detector: LicenseDetector = args.detector
    paths: Sequence[Path] = args.directory
    ignore_undetected: bool = args.ignore_undetected
    ignore_unlicensed_mods: bool = args.ignore_unlicensed_mods
    mode: str = args.mode
    verify: str | None = args.verify
    write_json: Path | None = args.write_json
    write_config: Path | None = args.write_config
    prompt: bool = args.prompt
    config_path: Path | None = args.config_path
    use_archive: bool = args.use_archive
    del args

    if write_config:
        config_path, loaded = get_report_write_config_data(config_path, detector)

    with handle_alternative_sources(paths, use_archive) as directory:
        license_data: LicenseData = detector.detect(directory)
        unlicensed_mods = (
            set()
            if ignore_unlicensed_mods
            else get_unlicensed_mods(directory, license_data.license_file_paths)
        )
        if prompt:
            license_data = write_and_prompt_report_licenses(license_data, loaded)
        print_licenses(
            license_data,
            unlicensed_mods,
            mode,
            not ignore_undetected,
            not ignore_unlicensed_mods,
            directory,
        )
    if write_json:
        write_license_json(license_data, write_json)
    if verify and not compare_licenses(license_data.license_expression, verify):
        sys.exit("Failed to verify license. Expected ^")
    if write_config:
        tomlkit_dump(loaded, cast(Path, config_path))
    sys.exit(
        bool(
            (license_data.undetected_licenses and not ignore_undetected)
            or (unlicensed_mods and not ignore_unlicensed_mods)
            or license_data.unmatched_extra_licenses
        )
    )


def _get_intermediate_directories(
    directory_parts: Collection[Sequence[str]],
) -> set[Path]:
    inter_parts = set()
    for parts in directory_parts:
        for i in range(len(parts)):
            inter_parts.add(Path(*parts[: i + 1]))
    return inter_parts


def copy_licenses(
    base_directory: Path,
    license_paths: Iterable[Path],
    install_destdir: Path,
    install_directory: Path,
    install_filelist: Path,
) -> None:
    installdir = install_destdir / install_directory.relative_to("/")
    base_directory = base_directory.resolve()

    entries: list[str] = []
    directory_parts: set[Sequence[str]] = set()
    installdir.mkdir(parents=True, exist_ok=True)
    entries.append(f"%license %dir {install_directory}")
    for lic in license_paths:
        resolvedpath = lic.resolve()
        relpath = resolvedpath.relative_to(base_directory)
        if len(relpath.parts) > 1:
            directory_parts.add(relpath.parts[:-1])
        (installdir / relpath).parent.mkdir(parents=True, exist_ok=True)
        shutil.copy2(lic, installdir / relpath)
        entries.append(f"%license {install_directory / relpath}")
    entries.extend(
        f"%license %dir {install_directory / path}"
        for path in _get_intermediate_directories(directory_parts)
    )
    entries.sort()
    install_filelist.write_text("\n".join(entries) + "\n")


def install_command(args: argparse.Namespace) -> None:
    """
    Install license files into the license directory
    """
    directory: Path = args.directory[0]
    detector: LicenseDetector = args.detector
    install_destdir: Path = args.install_destdir
    install_directory: Path = args.install_directory
    install_filelist: Path = args.install_filelist
    del args

    license_data: LicenseData = detector.detect(directory)
    copy_licenses(
        directory,
        sorted({*license_data.license_file_paths, *license_data.extra_license_files}),
        install_destdir,
        install_directory,
        install_filelist,
    )


def get_relpath(base_directory: Path, path: Path) -> Path:
    if path.is_absolute():
        return path.relative_to(base_directory)
    return path


def replace_entry(
    data: MutableSequence[LicenseEntry], new_entry: LicenseEntry, relpath: Path
) -> None:
    for entry in data:
        if entry == new_entry:
            return
        if Path(entry["path"]) == relpath:
            cast(dict, entry).clear()
            entry.update(new_entry)
            return
    data.append(new_entry)


def explicit_command(args: argparse.Namespace) -> None:
    if not args.config_path:
        sys.exit("--config must be specified!")
    loaded = load_tomlkit_if_exists(args.config_path)

    if "licensing" not in loaded:
        loaded.add("licensing", tomlkit.table())
    data = loaded["licensing"]

    licenses = cast(dict, data).setdefault("licenses", tomlkit.aot())
    relpath = get_relpath(args.directory[0], args.license_file)
    try:
        expression = (
            simplify_license(args.license_expression) if args.license_expression else ""
        )
    except license_expression.ExpressionError as exc:
        sys.exit(f"Failed to parse license: {exc}")
    entry = LicenseEntry(
        path=str(relpath),
        sha256sum=get_hash(args.license_file),
        expression=expression,
    )
    replace_entry(licenses, entry, relpath)
    tomlkit_dump(loaded, args.config_path)


def generate_buildrequires_command(args: argparse.Namespace) -> None:
    detector: str = args.detector_name
    del args

    if not detector:
        # If the detector is not explicitly specified, attempt to fall back to
        # the one whose dependencies are already installed.
        available, missing = get_detctors({}, create_license_config())
        detector = next(iter(available), "") or next(iter(missing))
    elif detector not in DETECTORS:
        sys.exit(f"{detector!r} is does not exist! Choices: {tuple(DETECTORS)}")
    detector_cls = DETECTORS[detector]
    for requirement in detector_cls.PACKAGES_NEEDED:
        print(requirement)


def main(argv: list[str] | None = None) -> None:
    args = parseargs(argv)
    if args.subcommand == "report":
        report_command(args)
    elif args.subcommand == "explicit":
        explicit_command(args)
    elif args.subcommand == "install":
        install_command(args)
    elif args.subcommand == "generate_buildrequires":
        generate_buildrequires_command(args)


if __name__ == "__main__":
    main()



# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

"""
Retrieve source entries from a specfile
"""

from __future__ import annotations

import re
import shutil
import subprocess
from collections.abc import Iterator
from pathlib import Path

from go_vendor_tools.exceptions import MissingDependencyError, VendorToolsError

SPECTOOL_PATH = shutil.which("spectool")

SOURCE_RE = re.compile(r"^Source(?P<number>\d+): (?P<source>.+)$", flags=re.MULTILINE)


def get_basename_heuristic(name: str) -> str:
    """
    Use a similar heuristic to that of spectool to determine a Source entry's
    basename
    """
    if "#" in name:
        _, basename = name.split("#")
        name = basename.lstrip("/")
    else:
        name = Path(name).name
    return name


def get_specfile_sources(spec_path: Path) -> Iterator[tuple[int, str]]:
    if not SPECTOOL_PATH:
        raise MissingDependencyError("spectool from rpmdevtools is missing!")
    stdout = subprocess.run(
        ["spectool", "--sources", spec_path], capture_output=True, text=True, check=True
    ).stdout.strip()
    for match in SOURCE_RE.finditer(stdout):
        yield int(match["number"]), match["source"]


def get_specfile_sources_relative(spec_path: Path) -> dict[int, str]:
    return {
        number: get_basename_heuristic(source)
        for number, source in get_specfile_sources(spec_path)
    }


def get_path_and_output_from_specfile(spec_path: Path) -> tuple[Path, Path]:
    sources = get_specfile_sources_relative(spec_path)
    if not {0, 1} & set(sources):
        # TODO(anyone): More specific exception class
        raise VendorToolsError(f"Source0 and Source1 must be specified in {spec_path}")
    directory = spec_path.resolve().parent
    return directory / sources[0], directory / sources[1]



# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

"""
Shared CLI utilities
"""

from __future__ import annotations

from pathlib import Path

import tomlkit


def load_tomlkit_if_exists(path: Path | None) -> tomlkit.TOMLDocument:
    if path and path.is_file():
        with path.open("r", encoding="utf-8") as fp:
            loaded = tomlkit.load(fp)
    else:
        loaded = tomlkit.document()
    return loaded



# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

"""
Exceptions used throughout the codebase
"""

from __future__ import annotations


class VendorToolsError(Exception):
    """
    Base Exception class
    """


class MissingDependencyError(VendorToolsError):
    """
    An optional dependency required by this operation is missing
    """


class LicenseError(VendorToolsError):
    """
    An issue occured while detecting licenses
    """


class ConfigError(VendorToolsError):
    """
    Failed to load config
    """


class ArchiveError(VendorToolsError):
    """
    An issue occured while creating an archive
    """



# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

"""
Utilities for reading configuartion
"""

from __future__ import annotations

import os

FALSY_STRINGS = frozenset(("", "0", "false"))


def get_envvar_boolean(variable: str, default: bool) -> bool:
    gotten = os.environ.get(variable)
    if gotten is None:
        return default
    return gotten.lower() not in FALSY_STRINGS



# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

"""
Utilities for working with license expressions
"""

from __future__ import annotations

import license_expression
from boolean.boolean import DualBase

licensing = license_expression.get_spdx_licensing()


def combine_licenses(
    *expressions: str | license_expression.LicenseExpression | None,
    validate=True,
    strict=True,
) -> license_expression.LicenseExpression:
    """
    Combine SPDX license expressions with AND
    """
    # Set a file's license to an empty string or None to exclude it from the
    # calculation.
    filtered = [str(expression) for expression in expressions if expression]
    filtered.sort()
    return simplify_license(
        str(license_expression.combine_expressions(filtered, licensing=licensing)),
        validate=validate,
        strict=strict,
    )


def _sort_expression_recursive(
    parsed: license_expression.LicenseExpression, /
) -> license_expression.LicenseExpression:
    if isinstance(parsed, DualBase) and (args := getattr(parsed, "args", None)):
        rec_sorted = sorted((_sort_expression_recursive(arg) for arg in args))
        parsed = parsed.__class__(*rec_sorted)
    return parsed


def simplify_license(
    expression: str | license_expression.LicenseExpression,
    *,
    validate: bool = True,
    strict: bool = True,
) -> str:
    """
    Simplify and verify a license expression
    """
    parsed = licensing.parse(str(expression), validate=validate, strict=strict)
    # DualBase subclasses are collections of licenses with an "AND" or an "OR"
    # relationship.
    if not isinstance(parsed, DualBase):
        return str(parsed)
    # Flatten licenses (e.g., "(MIT AND ISC) AND MIT" -> "MIT AND ISC"
    parsed = parsed.flatten()
    # Perform further license_expression-specific deduplication
    parsed = licensing.dedup(parsed)
    # Recursively sort AND/OR expressions
    parsed = _sort_expression_recursive(parsed)
    return str(parsed)


def compare_licenses(
    license1: license_expression.LicenseExpression | str,
    license2: str | license_expression.LicenseExpression | str,
    /,
) -> bool:
    return simplify_license(license1) == simplify_license(license2)



# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

"""
Detect licenses using trivy
"""

from __future__ import annotations

import dataclasses
import json
import shutil
import subprocess
from collections.abc import Sequence
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal, TypedDict, cast

from go_vendor_tools.config.licenses import LicenseConfig
from go_vendor_tools.licensing import combine_licenses

from .base import (
    LicenseData,
    LicenseDetector,
    LicenseDetectorNotAvailableError,
    filter_license_map,
    find_extra_license_files,
    get_extra_licenses,
)

if TYPE_CHECKING:
    from _typeshed import StrPath


class TrivyLicenseFileEntry(TypedDict):
    Severity: str
    Category: str
    PkgName: str
    FilePath: str
    Name: str
    Confidence: float
    Link: str


class TrivyLicenseDict(TypedDict):
    Target: Literal["Loose File License(s)"]
    Class: Literal["license-file"]
    Licenses: list[TrivyLicenseFileEntry]


def run_read_json(command: Sequence[StrPath]) -> Any:
    proc: subprocess.CompletedProcess[str] = subprocess.run(
        command, check=True, text=True, capture_output=True
    )
    return json.loads(proc.stdout)


@dataclasses.dataclass()
class TrivyLicenseData(LicenseData):
    trivy_license_data: TrivyLicenseDict


class TrivyLicenseDetector(LicenseDetector[TrivyLicenseData]):
    NAME = "trivy"
    PACKAGES_NEEDED = ("trivy",)

    def __init__(
        self, cli_config: dict[str, str], license_config: LicenseConfig
    ) -> None:
        if path := cli_config.get("trivy_path"):
            if not Path(path).exists():
                raise LicenseDetectorNotAvailableError(f"{path!r} does not exist!")
        else:
            path = shutil.which("trivy")
        if not path:
            raise LicenseDetectorNotAvailableError("Failed to find trivy binary!")

        self.path: str = path
        self.license_config = license_config

    # TODO(anyone): Consider splitting into separate functions
    # https://gitlab.com/gotmax23/go-vendor-tools/-/issues/23
    def detect(self, directory: StrPath) -> TrivyLicenseData:
        # fmt: off
        cmd = [
            self.path,
            "fs",
            "--scanners", "license",
            "--license-full",
            "-f", "json",
            directory,
        ]
        # fmt: on
        data = run_read_json(cmd)
        for item in data["Results"]:
            if item.get("Class") == "license-file":
                licenses = cast(TrivyLicenseDict, item)
                break
        else:
            raise ValueError("Failed to read Trivy license data")

        license_map: dict[Path, str] = {}
        for result in licenses.get("Licenses", []):
            path = Path(result["FilePath"])
            name = result["Name"]
            # License files can have multiple matches in trivy
            if path in license_map:
                license_map[path] = str(
                    combine_licenses(
                        # No need to validate. We do that later.
                        license_map[path],
                        name,
                        validate=False,
                        strict=False,
                    )
                )
            else:
                license_map[path] = name

        extra, unmatched = get_extra_licenses(
            self.license_config["licenses"], directory
        )
        license_map |= extra
        filtered_license_map = filter_license_map(
            license_map,
            self.license_config["exclude_directories"],
            self.license_config["exclude_files"],
        )
        filtered_license_map = dict(
            sorted(filtered_license_map.items(), key=lambda item: item[0])
        )

        extra_license_files = list(
            find_extra_license_files(
                directory,
                self.license_config["exclude_directories"],
                self.license_config["exclude_files"],
            )
        )
        return TrivyLicenseData(
            directory=Path(directory),
            license_map=filtered_license_map,
            # Trivy doesn't include undetected_licenses
            undetected_licenses=[],
            unmatched_extra_licenses=unmatched,
            trivy_license_data=licenses,
            extra_license_files=extra_license_files,
        )



# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

"""
Configuration for the go_vendor_licenses command
"""

from __future__ import annotations

import os
from typing import Any, TypedDict, cast


class LicenseEntry(TypedDict, total=False):
    path: str
    sha256sum: str
    expression: str


class LicenseConfig(TypedDict, total=False):
    """
    TypedDict representing the script's config file
    """

    detector: str | None
    licenses: list[LicenseEntry]
    exclude_directories: list[str]
    exclude_files: list[str]


def create_license_config(data: dict[str, Any] | None = None) -> LicenseConfig:
    data = {} if data is None else data.copy()
    data.setdefault("detector", os.environ.get("GO_VENDOR_LICENSE_DETECTOR"))
    data.setdefault("licenses", [])
    data.setdefault("exclude_globs", [])
    data.setdefault("exclude_directories", [])
    data.setdefault("exclude_files", [])
    return cast("LicenseConfig", data)



# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

"""
Base configuration
"""

from __future__ import annotations

import sys
from pathlib import Path
from typing import TYPE_CHECKING, Any, TypedDict, cast

from go_vendor_tools.config.archive import ArchiveConfig, create_archive_config

from .licenses import LicenseConfig, create_license_config

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib


if TYPE_CHECKING:
    from _typeshed import StrPath

FALSY_STRINGS = ("", "0", "false")


class BaseConfig(TypedDict):
    licensing: LicenseConfig
    archive: ArchiveConfig


def create_base_config(data: dict[str, Any] | None = None) -> BaseConfig:
    data = {} if data is None else data.copy()
    data["licensing"] = create_license_config(data.get("licensing"))
    data["archive"] = create_archive_config(data.get("archive"))
    return cast(BaseConfig, data)


def load_config(
    config: StrPath | None = None, allow_missing: bool = False
) -> BaseConfig:
    """
    Load the configuration TOML file if `config` is not None
    """
    if allow_missing and config and not Path(config).is_file():
        config = None
    if not config:
        return create_base_config()
    with open(config, "rb") as fp:
        data = tomllib.load(fp)
    return create_base_config(data)



# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

"""
Base classes for handling license detection tools
"""

from __future__ import annotations

import abc
import dataclasses
import os
import re
from collections.abc import Collection, Iterator
from itertools import chain
from pathlib import Path
from typing import TYPE_CHECKING, Any, ClassVar, Generic, TypeVar

import license_expression

from go_vendor_tools.config.licenses import LicenseConfig, LicenseEntry
from go_vendor_tools.exceptions import LicenseError
from go_vendor_tools.hashing import verify_hash
from go_vendor_tools.licensing import combine_licenses

if TYPE_CHECKING:
    from _typeshed import StrPath

EXTRA_LICENSE_FILE_REGEX = re.compile(r"^(AUTHORS|NOTICE).*$", flags=re.IGNORECASE)


def get_extra_licenses(
    licenses: list[LicenseEntry], directory: StrPath
) -> tuple[dict[Path, str], list[Path]]:
    results: dict[Path, str] = {}
    not_matched: list[Path] = []
    seen: set[Path] = set()
    for lic in licenses:
        relpath = Path(lic["path"])
        path = directory / relpath
        if path in results:
            raise LicenseError(
                f"{path} was specified multiple times in the configuration!"
            )
        seen.add(path)
        if verify_hash(path, lic["sha256sum"]):
            results[relpath] = lic["expression"]
        else:
            not_matched.append(relpath)
    return results, not_matched


def is_unwanted_path(
    path: Path,
    exclude_directories: Collection[StrPath],
    exclude_files: Collection[StrPath],
) -> bool:
    return (
        # Hardcoded exception
        "testdata" in path.parts
        or path in exclude_files
        or any(path.is_relative_to(directory) for directory in exclude_directories)
    )


def filter_license_map(
    license_map: dict[Path, str],
    exclude_directories: Collection[StrPath],
    exclude_files: Collection[StrPath],
) -> dict[Path, str]:
    """
    Filter licenses files from unwanted paths
    """
    exclude_directories = set(exclude_directories)
    exclude_files = {Path(file) for file in exclude_files}
    return {
        path: exp
        for path, exp in license_map.items()
        if not is_unwanted_path(path, exclude_directories, exclude_files)
    }


def find_extra_license_files(
    directory: StrPath,
    exclude_directories: Collection[StrPath],
    exclude_files: Collection[StrPath],
) -> Iterator[Path]:
    """
    Determine extra files (e.g., AUTHORS or NOTICE files) that we should
    include in the distribution but not run through the license detector
    """
    for root, _, files in os.walk(directory):
        for file in files:
            path = Path(root, file)
            relpath = path.relative_to(directory)
            if is_unwanted_path(relpath, exclude_directories, exclude_files):
                continue
            if EXTRA_LICENSE_FILE_REGEX.match(file):
                yield path


@dataclasses.dataclass()
class LicenseData:
    """
    Generic class representing detected license data.
    Can be subclassed by detector implementations to add additional fields.

    Attributes:
        directory:
            Path that was crawled for licensed
        license_map:
            Mapping of relative paths to license (within `directory`) to str
            SPDX license expressions
        undetected_licenses:
            License files that the license detector implementation failed to
            detect
        license_set:
            Set of unique detected license expressions
        license_expression:
            Cumulative `license_expression.LicenseExpression` SPDX expression
        license_files_paths:
            Full paths to all detected license files
        extra_license_files:
            Extra files (e.g., AUTHORS or NOTICE files) that we should include
            in the distribution but not run through the license detector
    """

    directory: Path
    license_map: dict[Path, str]
    undetected_licenses: Collection[Path]
    unmatched_extra_licenses: Collection[Path]
    license_set: set[str] = dataclasses.field(init=False)
    license_expression: license_expression.LicenseExpression | None = dataclasses.field(
        init=False
    )
    license_file_paths: Collection[Path] = dataclasses.field(init=False)
    extra_license_files: list[Path]
    _LIST_PATH_FIELDS: ClassVar = (
        "undetected_licenses",
        "unmatched_extra_licenses",
        "license_file_paths",
        "extra_license_files",
    )
    replace = dataclasses.replace

    def __post_init__(self) -> None:
        self.license_set = set(self.license_map.values())
        self.license_expression = (
            combine_licenses(*self.license_set) if self.license_map else None
        )
        self.license_file_paths = tuple(
            self.directory / lic
            for lic in chain(self.license_map, self.undetected_licenses)
        )

    # TODO(gotmax23): Consider cattrs or pydantic
    def to_jsonable(self) -> dict[str, Any]:
        data = dataclasses.asdict(self)
        for key, value in data.items():
            if key == "directory":
                data[key] = str(value)
            elif key == "license_map":
                data[key] = {str(key1): value1 for key1, value1 in value.items()}
            elif key in self._LIST_PATH_FIELDS:
                data[key] = list(map(str, value))
            elif key == "license_set":
                data[key] = list(value)
            elif key == "license_expression":
                data[key] = str(value)
        return data

    @classmethod
    def _from_jsonable_to_dict(cls, data: dict[Any, Any]) -> dict[Any, Any]:
        init_fields = [field.name for field in dataclasses.fields(cls) if field.init]
        newdata: dict[Any, Any] = {}
        for key, value in data.items():
            if key not in init_fields:
                continue
            if key == "directory":
                newdata[key] = Path(value)
            elif key == "license_map":
                newdata[key] = {Path(key1): value1 for key1, value1 in value.items()}
            elif key in cls._LIST_PATH_FIELDS:
                newdata[key] = tuple(map(Path, value))
            else:
                newdata[key] = value
        return newdata

    @classmethod
    def from_jsonable(cls: type[LicenseDataT], data: dict[Any, Any]) -> LicenseDataT:
        return cls(**cls._from_jsonable_to_dict(data))


LicenseDataT = TypeVar("LicenseDataT", bound=LicenseData)


class LicenseDetector(Generic[LicenseDataT], metaclass=abc.ABCMeta):
    NAME: ClassVar[str]
    PACKAGES_NEEDED: ClassVar[tuple[str, ...]] = ()

    @abc.abstractmethod
    def __init__(
        self, cli_config: dict[str, str], license_config: LicenseConfig
    ) -> None: ...
    @abc.abstractmethod
    def detect(self, directory: StrPath) -> LicenseDataT: ...


class LicenseDetectorNotAvailableError(LicenseError):
    """
    Failed to load the requested license detector
    """



# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

"""
Parse go.mod and modules.txt files
"""

from __future__ import annotations

import re
from collections.abc import Collection
from pathlib import Path

MODULE_REGEX = re.compile(r"^# (?:.+=>)?(?P<ipath>\S+) v(?P<version>\S+)$")


def get_go_module_names(directory: Path) -> dict[str, str]:
    results: dict[str, str] = {}
    with (directory / "vendor/modules.txt").open("r", encoding="utf-8") as fp:
        for line in fp:
            if match := MODULE_REGEX.match(line):
                results[match["ipath"]] = match["version"]
    return results


def get_go_module_dirs(directory: Path) -> list[Path]:
    results: list[Path] = []
    for ipath in get_go_module_names(directory):
        moddir = directory / "vendor" / ipath
        if moddir.is_dir():
            results.append(moddir.resolve())
    return results


def get_unlicensed_mods(directory: Path, license_paths: Collection[Path]) -> set[Path]:
    resolved_dir = directory.resolve()
    licensed_dirs = {
        (
            first.parent
            if (first := path.parent).relative_to(resolved_dir).name == "LICENSES"
            else first
        )
        for path in (p.resolve() for p in license_paths)
    }
    all_dirs = {*get_go_module_dirs(directory), directory.resolve()}
    return all_dirs - licensed_dirs



# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

"""
Configuration for the go_vendor_archive command
"""

from __future__ import annotations

import os
from collections.abc import Iterator
from typing import Any, Mapping, TypedDict, cast

from .utils import get_envvar_boolean

DEFAULT_USE_TOP_LEVEL_DIR = False
DEFAULT_USE_MODULE_PROXY = get_envvar_boolean(
    "GO_VENDOR_ARCHIVE_USE_MODULE_PROXY", True
)
DEFAULT_TIDY = True
DEFAULT_COMPRESSLEVEL_STR = os.environ.get("GO_VENDOR_ARCHIVE_COMPRESSLEVEL")


class ArchiveConfig(TypedDict):
    use_module_proxy: bool
    use_top_level_dir: bool
    # Commands to run before downloading modules
    pre_commands: list[list[str]]
    # Commands to run after downloading modules
    post_commands: list[list[str]]
    tidy: bool
    dependency_overrides: dict[str, str]
    compresslevel: int | None
    compression_type: str | None


def create_archive_config(config: dict[str, Any] | None = None) -> ArchiveConfig:
    # Keep the same order here as in ArchiveConfig definition
    config = {} if config is None else config.copy()
    config.setdefault("use_module_proxy", DEFAULT_USE_MODULE_PROXY)
    config.setdefault("use_top_level_dir", DEFAULT_USE_TOP_LEVEL_DIR)
    config.setdefault("pre_commands", [])
    config.setdefault("post_commands", [])
    config.setdefault("tidy", DEFAULT_TIDY)
    config.setdefault("dependency_overrides", {})
    config.setdefault(
        "compresslevel",
        DEFAULT_COMPRESSLEVEL_STR if DEFAULT_COMPRESSLEVEL_STR else None,
    )
    if config["compresslevel"] is not None:
        config["compresslevel"] = int(config["compresslevel"])
    config.setdefault("compression_type", None)
    return cast(ArchiveConfig, config)


def get_go_dependency_update_commands(
    dependency_overrides: Mapping[str, str]
) -> Iterator[tuple[str, ...]]:
    for ipath, version in dependency_overrides.items():
        yield ("go", "get", f"{ipath}@{version}")



# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

"""
SHA256 hashing utilities
"""

from __future__ import annotations

import hashlib
import hmac
from pathlib import Path


def get_hash(file: Path) -> str:
    hasher = hashlib.sha256()
    with file.open("rb") as fp:
        while chunk := fp.read(4096):
            hasher.update(chunk)
    return hasher.hexdigest()


def verify_hash(file: Path, sha256sum: str) -> bool:
    if not file.is_file():
        return False
    return hmac.compare_digest(get_hash(file), sha256sum)



# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

from __future__ import annotations

import sys
from pathlib import Path

from pytest_mock import MockerFixture

from go_vendor_tools.cli.go_vendor_archive import (
    CreateArchiveArgs,
    OverrideArgs,
    create_archive,
    override_command,
)
from go_vendor_tools.config.base import create_base_config, load_config

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomlkit as tomllib


def test_vendor_archive_base(mocker: MockerFixture, tmp_path: Path) -> None:
    patched_run_command = mocker.patch(
        "go_vendor_tools.cli.go_vendor_archive.run_command"
    )
    mocked_add_files_to_archive = mocker.patch(  # noqa F841
        "go_vendor_tools.cli.go_vendor_archive.add_files_to_archive"
    )
    (directory := tmp_path / "directory").mkdir()
    config = create_base_config(
        {
            "archive": {
                "dependency_overrides": {"golang.org/x/sys": "v0.6.0"},
                "post_commands": [["echo", "Hello world!"]],
            }
        }
    )
    args = CreateArchiveArgs(
        path=directory,
        output=Path("vendor.tar.xz"),
        use_top_level_dir=False,
        use_module_proxy=False,
        tidy=True,
        idempotent=False,
        compression_type=None,
        compresslevel=None,
        config_path=tmp_path / "go-vendor-tools.toml",
        config=config,
        write_config=False,
    )
    create_archive(args)
    expected_calls = [
        ["go", "get", "golang.org/x/sys@v0.6.0"],
        ["go", "mod", "tidy"],
        ["go", "mod", "vendor"],
        ["echo", "Hello world!"],
    ]
    calls = [list(c.args[1]) for c in patched_run_command.call_args_list]
    assert expected_calls == calls


def test_vendor_archive_write_config(tmp_path: Path) -> None:
    args = CreateArchiveArgs.construct(
        subcommand="create",
        path=tmp_path,
        output=Path("vendor.tar.bz2"),
        idempotent=False,
        compression_type="bz2",
        compresslevel=9,
        config_path=tmp_path / "go-vendor-tools.toml",
        write_config=True,
    )
    args.write_config_opts()
    with open(args.config_path, "rb") as fp:
        data = tomllib.load(fp)
    assert data == {"archive": {"compression_type": "bz2", "compresslevel": 9}}


def test_vendor_archive_override(tmp_path: Path) -> None:
    config_path = tmp_path / "config.toml"
    args = OverrideArgs(
        config_path=config_path, import_path="golang.org/x/sys", version="v0.6.0"
    )
    override_command(args)
    output_config = load_config(config_path)
    expected_dependency_overrides = {"golang.org/x/sys": "v0.6.0"}
    gotten_dependency_overrides = output_config["archive"]["dependency_overrides"]
    assert expected_dependency_overrides == gotten_dependency_overrides



# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

from __future__ import annotations

import sys
from pathlib import Path
from shutil import copy2

import pytest

from go_vendor_tools.cli import go_vendor_license
from go_vendor_tools.config.base import load_config
from go_vendor_tools.license_detection.base import (
    LicenseData,
    LicenseDetector,
    get_extra_licenses,
)
from go_vendor_tools.license_detection.load import get_detctors

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib

HERE = Path(__file__).resolve().parent
TEST_DATA = HERE / "test_data"

CONFIG1 = load_config(TEST_DATA / "case1" / "config.toml")
CONFIG1_BROKEN = load_config(TEST_DATA / "case1" / "config-broken.toml")


def get_available_detectors() -> list[type[LicenseDetector]]:
    # TODO(anyone): Allow enforcing "strict mode" if any detectors are missing
    # This can be a env var and then enabled in the noxfile.
    available, missing = get_detctors({}, CONFIG1["licensing"])
    # HACK: We initialize the classes using a test config to check if they are
    # available and then return the base class so that it can be reinitialized
    return [type(d) for d in available.values()]


@pytest.fixture(name="detector", params=get_available_detectors())
def get_detectors(request) -> type[LicenseDetector]:
    return request.param


def test_license_explicit(test_data: Path, tmp_path: Path) -> None:
    case_dir = test_data / "case1"
    licenses_dir = case_dir / "licenses"
    with open(case_dir / "config.toml", "rb") as fp:
        expected = tomllib.load(fp)
    dest = tmp_path / "config.toml"
    copy2(case_dir / "config-broken.toml", dest)
    go_vendor_license.main(
        [
            f"-c{dest}",
            f"-C{licenses_dir}",
            "explicit",
            f"-f{licenses_dir / 'LICENSE.MIT'}",
            "MIT",
        ]
    )
    with open(dest, "rb") as fp:
        gotten = tomllib.load(fp)
    assert gotten == expected


def test_get_extra_licenses(test_data: Path) -> None:
    case_dir = test_data / "case1"
    licenses_dir = case_dir / "licenses"
    config = load_config(case_dir / "config.toml")
    matched, missing = get_extra_licenses(config["licensing"]["licenses"], licenses_dir)
    expected_map = {
        Path("LICENSE.BSD3"): "BSD-3-Clause",
        Path("LICENSE.MIT"): "MIT",
    }
    assert matched == expected_map
    assert not missing


def test_get_extra_licenses_error(test_data: Path) -> None:
    case_dir = test_data / "case1"
    licenses_dir = case_dir / "licenses"
    matched, missing = get_extra_licenses(
        CONFIG1_BROKEN["licensing"]["licenses"], licenses_dir
    )
    expected_map = {Path("LICENSE.BSD3"): "BSD-3-Clause"}
    assert matched == expected_map
    assert missing == [Path("LICENSE.MIT")]


def test_load_dump_license_data(
    test_data: Path, detector: type[LicenseDetector]
) -> None:
    case_dir = test_data / "case2"
    licenses_dir = case_dir / "licenses"
    config = load_config(None)
    detector_obj = detector({}, config["licensing"])
    data: LicenseData = detector_obj.detect(licenses_dir)
    jsonable = data.to_jsonable()
    new_data = type(data).from_jsonable(jsonable)
    assert new_data.to_jsonable() == jsonable


def test_detect_nothing(tmp_path: Path, detector: type[LicenseDetector]) -> None:
    """
    Ensure the code has proper error handling for when no licenses are detected
    """
    config = load_config(None)
    detector_obj = detector({}, config["licensing"])
    data: LicenseData = detector_obj.detect(tmp_path)
    assert data.directory == tmp_path
    assert not data.license_map
    assert not data.undetected_licenses
    assert not data.license_set
    assert data.license_expression is None



# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

from __future__ import annotations

import pytest

from go_vendor_tools.licensing import (
    combine_licenses,
    compare_licenses,
    simplify_license,
)


@pytest.mark.parametrize(
    "expression, expected",
    [
        pytest.param("GPL-2.0-only", "GPL-2.0-only"),
        pytest.param("GPL-2.0-only OR GPL-2.0", "GPL-2.0-only"),
        pytest.param(
            "(MIT OR Apache-2.0) AND Apache-2.0",
            "Apache-2.0 AND (Apache-2.0 OR MIT)",
        ),
        pytest.param(
            "(Apache-2.0 OR MIT) AND Apache-2.0",
            "Apache-2.0 AND (Apache-2.0 OR MIT)",
        ),
        pytest.param(
            "(Apache-2.0 AND (MIT OR Apache-2.0)) AND Apache-2.0",
            "Apache-2.0 AND (Apache-2.0 OR MIT)",
        ),
    ],
)
def test_simplify_license(expression: str, expected: str) -> None:
    simplified = simplify_license(expression)
    assert str(simplified) == expected
    assert compare_licenses(expected, simplified)


def test_combine_licenses() -> None:
    combined = combine_licenses("Apache-2.0 OR MIT", "ISC")
    assert str(combined) == "ISC AND (Apache-2.0 OR MIT)"



#!/usr/bin/env python3
# PYTHON_ARGCOMPLETE_OK
# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

from __future__ import annotations

import argparse
import dataclasses
import os
import shlex
import shutil
import subprocess
import sys
import tempfile
from collections.abc import Callable, Sequence
from contextlib import ExitStack
from functools import partial
from itertools import chain
from pathlib import Path
from typing import TYPE_CHECKING, Any, ClassVar

from zstarfile import ZSTarfile
from zstarfile.extra import open_write_compressed

from go_vendor_tools import __version__
from go_vendor_tools.archive import add_files_to_archive
from go_vendor_tools.config.archive import get_go_dependency_update_commands
from go_vendor_tools.config.base import BaseConfig, load_config
from go_vendor_tools.exceptions import ArchiveError
from go_vendor_tools.specfile_sources import get_path_and_output_from_specfile

try:
    import tomlkit
except ImportError:
    HAS_TOMLKIT = False
else:
    HAS_TOMLKIT = True
    from go_vendor_tools.cli.utils import load_tomlkit_if_exists

try:
    import argcomplete
except ImportError:
    HAS_ARGCOMPLETE = False
else:
    HAS_ARGCOMPLETE = True

if TYPE_CHECKING:
    from _typeshed import StrPath

DEFAULT_OUTPUT = "vendor.tar.bz2"
ARCHIVE_FILES = (Path("go.mod"), Path("go.sum"), Path("vendor"))
OPTIONAL_FILES = frozenset({Path("go.sum")})
GO_PROXY_ENV = {
    "GOPROXY": "https://proxy.golang.org,direct",
    "GOSUMDB": "sum.golang.org",
}


def need_tomlkit(action="this action"):
    if not HAS_TOMLKIT:
        message = f"tomlkit is required for {action}. Please install it!"
        sys.exit(message)


def run_command(
    runner: Callable[..., subprocess.CompletedProcess],
    command: Sequence[StrPath],
    **kwargs: Any,
) -> subprocess.CompletedProcess:
    print(f"$ {shlex.join(map(os.fspath, command))}")  # type: ignore[arg-type]
    return runner(command, **kwargs)


@dataclasses.dataclass()
class CreateArchiveArgs:
    path: Path
    output: Path
    use_top_level_dir: bool
    use_module_proxy: bool
    tidy: bool
    idempotent: bool
    compresslevel: int | None
    compression_type: str | None
    config_path: Path
    config: BaseConfig
    write_config: bool
    _explicitly_passed: list[str] = dataclasses.field(default_factory=list, repr=False)

    CONFIG_OPTS: ClassVar[tuple[str, ...]] = (
        "use_module_proxy",
        "use_top_level_dir",
        "tidy",
        "compresslevel",
        "compression_type",
    )

    @classmethod
    def construct(cls, **kwargs: Any) -> CreateArchiveArgs:
        if kwargs.pop("subcommand") != "create":
            raise AssertionError  # pragma: no cover
        _explicitly_passed = list(kwargs)
        kwargs["config"] = load_config(kwargs["config_path"], kwargs["write_config"])
        for opt in cls.CONFIG_OPTS:
            if kwargs.get(opt) is None:
                kwargs[opt] = kwargs["config"]["archive"][opt]
        if not kwargs["path"].exists():
            raise ArchiveError(f"{kwargs['path']} does not exist!")
        if kwargs["write_config"]:
            need_tomlkit("--write-config")
            if not kwargs["config_path"]:
                raise ArchiveError("--write-config requires --config to be set")
        return CreateArchiveArgs(**kwargs, _explicitly_passed=_explicitly_passed)

    def write_config_opts(self) -> None:
        need_tomlkit("write_config_opts")
        loaded = load_tomlkit_if_exists(self.config_path)
        config = loaded.setdefault("archive", {})
        for opt in self._explicitly_passed:
            if opt not in self.CONFIG_OPTS:
                continue
            config[opt] = getattr(self, opt)
        with open(self.config_path, "w", encoding="utf-8") as fp:
            tomlkit.dump(loaded, fp)


@dataclasses.dataclass()
class OverrideArgs:
    config_path: Path
    import_path: str
    version: str

    @classmethod
    def construct(cls, **kwargs: Any) -> OverrideArgs:
        if kwargs.pop("subcommand") != "override":
            raise AssertionError  # pragma: no cover
        return cls(**kwargs)


def parseargs(argv: list[str] | None = None) -> CreateArchiveArgs | OverrideArgs:
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(dest="subcommand")
    subparsers.required = True
    create_subparser = subparsers.add_parser("create")
    create_subparser.add_argument("--version", action="version", version=__version__)
    create_subparser.add_argument(
        "-O", "--output", type=Path, default=None, help=f"Default: {DEFAULT_OUTPUT}"
    )
    create_subparser.add_argument(
        "--top-level-dir",
        dest="use_top_level_dir",
        action=argparse.BooleanOptionalAction,
        default=argparse.SUPPRESS,
    )
    create_subparser.add_argument(
        "--use-module-proxy",
        action=argparse.BooleanOptionalAction,
        default=argparse.SUPPRESS,
    )
    create_subparser.add_argument(
        "-p", action="store_true", dest="use_module_proxy", default=argparse.SUPPRESS
    )
    create_subparser.add_argument("-c", "--config", type=Path, dest="config_path")
    create_subparser.add_argument(
        "--tidy",
        action=argparse.BooleanOptionalAction,
        default=argparse.SUPPRESS,
    )
    create_subparser.add_argument(
        "-I",
        "--idempotent",
        action="store_true",
        help="Only generate archive if OUTPUT does not already exist",
    )
    create_subparser.add_argument(
        "--compresslevel", type=int, default=argparse.SUPPRESS
    )
    create_subparser.add_argument(
        "--compression",
        dest="compression_type",
        metavar="COMPRESSION_TYPE",
        help=f"Choices: {list(ZSTarfile.OPEN_METH)}",
        default=argparse.SUPPRESS,
    )
    create_subparser.add_argument("--write-config", action="store_true")
    create_subparser.add_argument("path", type=Path)
    override_subparser = subparsers.add_parser("override")
    override_subparser.add_argument(
        "--config", type=Path, dest="config_path", required=True
    )
    override_subparser.add_argument("import_path")
    override_subparser.add_argument("version")

    if HAS_ARGCOMPLETE:
        argcomplete.autocomplete(parser)
    args = parser.parse_args(argv)
    if args.subcommand == "create":
        return CreateArchiveArgs.construct(**vars(args))
    elif args.subcommand == "override":
        return OverrideArgs.construct(**vars(args))
    else:
        raise RuntimeError("unreachable")


def _create_archive_read_from_specfile(args: CreateArchiveArgs) -> None:
    if args.output:
        sys.exit("Cannot pass --output when reading paths from a specfile!")
    spec_path = args.path
    args.path, args.output = get_path_and_output_from_specfile(args.path)
    if not args.path.is_file():
        sys.exit(
            f"{args.path} does not exist!"
            f" Run 'spectool -g {spec_path}' and try again!"
        )


def create_archive(args: CreateArchiveArgs) -> None:
    _already_checked_is_file = False
    cwd = args.path
    if args.path.suffix == ".spec":
        _create_archive_read_from_specfile(args)
        _already_checked_is_file = True
    elif not args.output:
        args.output = Path(DEFAULT_OUTPUT)
    if args.idempotent and args.output.exists():
        print(f"{args.output} already exists")
        sys.exit()
    with ExitStack() as stack:
        try:
            tf = stack.enter_context(
                open_write_compressed(
                    args.output,
                    compression_type=args.compression_type,
                    compresslevel=args.compresslevel,
                )
            )
        except ValueError as exc:
            sys.exit(f"Invalid --output value: {exc}")
        # Treat as an archive if it's not a directory
        if _already_checked_is_file or args.path.is_file():
            print(f"* Treating {args.path} as an archive. Unpacking...")
            cwd = Path(stack.enter_context(tempfile.TemporaryDirectory()))
            shutil.unpack_archive(args.path, cwd)
            cwd /= next(cwd.iterdir())
        env = os.environ | GO_PROXY_ENV if args.use_module_proxy else None
        runner = partial(subprocess.run, cwd=cwd, check=True, env=env)
        pre_commands = chain(
            args.config["archive"]["pre_commands"],
            get_go_dependency_update_commands(
                args.config["archive"]["dependency_overrides"]
            ),
        )
        for command in pre_commands:
            run_command(runner, command)
        if args.tidy:
            run_command(runner, ["go", "mod", "tidy"])
        run_command(runner, ["go", "mod", "vendor"])
        # Create vendor directory so it is there even if there are no
        # dependencies to download
        (vdir := cwd / "vendor").mkdir(exist_ok=True)
        (vdir / "modules.txt").touch(exist_ok=True)
        for command in args.config["archive"]["post_commands"]:
            run_command(runner, command)
        print("Creating archive...")
        add_files_to_archive(
            tf,
            Path(cwd),
            ARCHIVE_FILES,
            top_level_dir=args.use_top_level_dir,
            optional_files=OPTIONAL_FILES,
        )
        if args.write_config:
            args.write_config_opts()


def override_command(args: OverrideArgs) -> None:
    need_tomlkit()
    loaded = load_tomlkit_if_exists(args.config_path)
    overrides = loaded.setdefault("archive", {}).setdefault("dependency_overrides", {})
    overrides[args.import_path] = args.version
    with open(args.config_path, "w", encoding="utf-8") as fp:
        tomlkit.dump(loaded, fp)


def main(argv: list[str] | None = None) -> None:
    args = parseargs(argv)
    if isinstance(args, CreateArchiveArgs):
        create_archive(args)
    elif isinstance(args, OverrideArgs):
        override_command(args)


if __name__ == "__main__":
    try:
        main()
    except ArchiveError as exc:
        sys.exit(str(exc))



# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

from __future__ import annotations

from pathlib import Path

import pytest

HERE = Path(__file__).resolve().parent
TEST_DATA = HERE / "test_data"


@pytest.fixture
def test_data() -> Path:
    return TEST_DATA



# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

"""
Load license detectors
"""

from __future__ import annotations

from ..config.licenses import LicenseConfig
from .askalono import AskalonoLicenseDetector
from .base import LicenseDetector, LicenseDetectorNotAvailableError
from .trivy import TrivyLicenseDetector

DETECTORS: dict[str, type[LicenseDetector]] = {
    TrivyLicenseDetector.NAME: TrivyLicenseDetector,
    AskalonoLicenseDetector.NAME: AskalonoLicenseDetector,
}


def get_detctors(
    cli_config: dict[str, str],
    license_config: LicenseConfig,
    detectors: dict[str, type[LicenseDetector]] = DETECTORS,
) -> tuple[dict[str, LicenseDetector], dict[str, LicenseDetectorNotAvailableError]]:
    found: dict[str, LicenseDetector] = {}
    errored: dict[str, LicenseDetectorNotAvailableError] = {}
    for name, class_ in detectors.items():
        try:
            detector = class_(cli_config, license_config)
        except LicenseDetectorNotAvailableError as exc:  # noqa PERF203
            errored[name] = exc
        else:
            found[name] = detector
    return found, errored



# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

"""
Detect licenses using askalono
"""

from __future__ import annotations

import dataclasses
import json
import shutil
import subprocess
from collections.abc import Callable
from pathlib import Path
from typing import TYPE_CHECKING, Any, TypedDict, cast

from ..config.licenses import LicenseConfig
from .base import (
    LicenseData,
    LicenseDetector,
    LicenseDetectorNotAvailableError,
    filter_license_map,
    find_extra_license_files,
    get_extra_licenses,
    is_unwanted_path,
)

if TYPE_CHECKING:
    from _typeshed import StrPath


class AskalonoLicenseEntry(TypedDict):
    name: str
    kind: str
    aliases: list[str]


class AskalonoLicenseResult(TypedDict):
    score: float
    license: AskalonoLicenseEntry
    containing: list[Any]


class AskalonoLicenseDict(TypedDict):
    path: str
    result: AskalonoLicenseResult


def _remove_line(file: StrPath, key: Callable[[str], bool]) -> None:
    """
    Used to remove vendor directory from .gitignore to avoid confusing askalono
    """
    lines: list[str] = []
    with open(file, "r+", encoding="utf-8") as fp:
        for line in fp:
            if key(line):
                continue
            lines.append(line)
        fp.seek(0)
        fp.writelines(lines)
        fp.truncate()


def _get_askalono_data(directory: StrPath) -> list[AskalonoLicenseDict]:
    """
    Crawl `directory` with askalono and return the serialized JSON representation
    """
    licenses_json = subprocess.run(
        ("askalono", "--format=json", "crawl", directory),
        check=True,
        capture_output=True,
    ).stdout.decode("utf-8")
    licenses = [
        cast(AskalonoLicenseDict, json.loads(line))
        for line in licenses_json.splitlines()
    ]
    return licenses


def _get_relative(base_dir: Path, file: str | Path) -> Path:
    file = Path(file)
    return file.relative_to(base_dir) if file.is_absolute() else file


def _filter_license_data(
    data: list[AskalonoLicenseDict],
    directory: Path,
) -> tuple[list[AskalonoLicenseDict], set[Path]]:

    undetected_licenses: set[Path] = set()
    results: list[AskalonoLicenseDict] = []

    for licensed in data:
        if "/PATENTS" not in licensed["path"] and "/NOTICE" not in licensed["path"]:
            try:
                licensed["result"]["license"]["name"]
            except KeyError:
                undetected_licenses.add(_get_relative(directory, licensed["path"]))
            else:
                results.append(licensed)
    return results, undetected_licenses


def _get_simplified_license_map(
    directory: Path,
    filtered_license_data: list[AskalonoLicenseDict],
    extra_license_mapping: dict[Path, str] | None = None,
) -> dict[Path, str]:
    """
    Given license data from askalono, return a simple mapping of license file
    Path to the license expression
    """
    results: dict[Path, str] = {}
    for licensed in filtered_license_data:
        license_name = licensed["result"]["license"]["name"]
        results[_get_relative(directory, licensed["path"])] = license_name
    results.update(extra_license_mapping or {})
    return dict(sorted(results.items(), key=lambda item: item[0]))


@dataclasses.dataclass()
class AskalonoLicenseData(LicenseData):
    askalono_license_data: list[AskalonoLicenseDict]


class AskalonoLicenseDetector(LicenseDetector[AskalonoLicenseData]):
    NAME = "askalono"
    PACKAGES_NEEDED = ("askalono-cli",)

    def __init__(
        self, cli_config: dict[str, str], license_config: LicenseConfig
    ) -> None:
        if path := cli_config.get("askalono_path"):
            if not Path(path).exists():
                raise LicenseDetectorNotAvailableError(f"{path!r} does not exist!")
        else:
            path = shutil.which("askalono")
        if not path:
            raise LicenseDetectorNotAvailableError("Failed to find askalono binary!")

        self.path: str = path
        self.license_config = license_config

    def detect(self, directory: StrPath) -> AskalonoLicenseData:
        gitignore = Path(directory, ".gitignore")
        if gitignore.is_file():
            _remove_line(gitignore, lambda line: line.startswith("vendor"))
        results, undetected = _filter_license_data(
            _get_askalono_data(directory), Path(directory)
        )
        extra_licenses, unmatched = get_extra_licenses(
            self.license_config["licenses"], directory
        )
        license_map = _get_simplified_license_map(
            Path(directory), results, extra_licenses
        )
        # Remove manually specified licenses
        undetected -= set(extra_licenses)
        undetected = {
            path
            for path in undetected
            if not is_unwanted_path(
                path,
                self.license_config["exclude_directories"],
                self.license_config["exclude_files"],
            )
        }

        filtered_license_map = filter_license_map(
            license_map,
            self.license_config["exclude_directories"],
            self.license_config["exclude_files"],
        )
        filtered_license_map = dict(
            sorted(filtered_license_map.items(), key=lambda item: item[0])
        )
        extra_license_files = list(
            find_extra_license_files(
                directory,
                self.license_config["exclude_directories"],
                self.license_config["exclude_files"],
            )
        )
        return AskalonoLicenseData(
            directory=Path(directory),
            license_map=filtered_license_map,
            undetected_licenses=undetected,
            unmatched_extra_licenses=unmatched,
            askalono_license_data=results,
            extra_license_files=extra_license_files,
        )



# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

"""
Tools for creating tar archives
"""

from __future__ import annotations

import os
import tarfile
from collections.abc import Collection, Sequence
from pathlib import Path, PurePath


### normalize_file_permissions is derived from flit_core
# Copyright (c) 2015, Thomas Kluyver and contributors
# SPDX-License-Identifier: BSD-3-Clause
def normalize_file_permissions(st_mode: int) -> int:
    """Normalize the permission bits in the st_mode field from stat to 644/755

    Popular VCSs only track whether a file is executable or not. The exact
    permissions can vary on systems with different umasks. Normalising
    to 644 (non executable) or 755 (executable) makes builds more reproducible.
    """
    # Set 644 permissions, leaving higher bits of st_mode unchanged
    new_mode = (st_mode | 0o644) & ~0o133
    if st_mode & 0o100:
        new_mode |= 0o111  # Executable: 644 -> 755
    return new_mode


def reproducible_filter(member: tarfile.TarInfo) -> tarfile.TarInfo:
    return member.replace(
        mtime=int(os.environ.get("SOURCE_DATE_EPOCH", 0)),
        mode=normalize_file_permissions(member.mode),
        uid=0,
        gid=0,
        uname="",
        gname="",
        deep=False,
    )


def add_files_to_archive(
    tf: tarfile.TarFile,
    directory: Path,
    files: Sequence[Path],
    top_level_dir: bool = False,
    optional_files: Collection[Path] = frozenset(),
) -> None:
    for file in files:
        if file.is_absolute():
            raise ValueError(f"{file} is not a relative path!")
        if not (directory / file).exists() and file in optional_files:
            continue
        dest = (directory.resolve().name / file) if top_level_dir else file
        tf.add(directory / file, dest, filter=reproducible_filter)


def get_toplevel_directory(tar: tarfile.TarFile) -> str | None:
    first_parent = {PurePath(info.name).parts[0] for info in tar.getmembers()}
    if len(first_parent) == 1:
        return next(iter(first_parent))
    return None



# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: MIT

from __future__ import annotations

import contextlib
import os
import shlex
from collections.abc import Iterator
from glob import iglob
from pathlib import Path
from tempfile import TemporaryDirectory
from types import SimpleNamespace

import nox

IN_CI = "JOB_ID" in os.environ or "CI" in os.environ
ALLOW_EDITABLE = os.environ.get("ALLOW_EDITABLE", str(not IN_CI)).lower() in (
    "1",
    "true",
)

PROJECT = "go_vendor_tools"
SPECFILE = "go-vendor-tools.spec"
LINT_SESSIONS = ("formatters", "codeqa", "typing")
LINT_FILES = (f"src/{PROJECT}", "tests/pytests", "noxfile.py")
INTEGRATION_PACKAGES = ("autorestic", "fzf")
COVERAGE_FAIL_UNDER = os.environ.get("COVERAGE_FAIL_UNDER") or "85"

nox.options.sessions = ("lint", "covtest")


# Helpers


def install(session: nox.Session, *args, editable=False, **kwargs):
    if editable and ALLOW_EDITABLE:
        args = ("-e", *args)
    session.install(*args, **kwargs)


def git(session: nox.Session, *args, **kwargs):
    return session.run("git", *args, **kwargs, external=True)


BASE_COVERAGE_COMMAND = ("coverage", "run", "-p", "--source", "go_vendor_tools")
COVERAGE_COMMANDS = SimpleNamespace(
    go_vendor_archive=(
        *BASE_COVERAGE_COMMAND,
        "-m",
        "go_vendor_tools.cli.go_vendor_archive",
    ),
    go_vendor_license=(
        *BASE_COVERAGE_COMMAND,
        "-m",
        "go_vendor_tools.cli.go_vendor_license",
    ),
)


@contextlib.contextmanager
def coverage_run(session: nox.Session) -> Iterator[dict[str, str]]:
    tmp = Path(session.create_tmp())
    covfile = (tmp / ".coverage").resolve()
    cov_env = {
        "COVERAGE_FILE": str(covfile),
        "GO_VENDOR_ARCHIVE": shlex.join(COVERAGE_COMMANDS.go_vendor_archive),
        "GO_VENDOR_LICENSE": shlex.join(COVERAGE_COMMANDS.go_vendor_license),
        **session.env,
    }
    yield cov_env
    combined = map(str, tmp.glob(".coverage.*"))
    session.run("coverage", "combine", *combined, env=cov_env)


# General


@nox.session(python=["3.9", "3.10", "3.11", "3.12"])
def test(session: nox.Session):
    packages: list[str] = [".[test]"]
    env: dict[str, str] = {}
    tmp = Path(session.create_tmp())

    if any(i.startswith("--cov") for i in session.posargs):
        packages.extend(("coverage[toml]", "pytest-cov"))
        env["COVERAGE_FILE"] = str(tmp / ".coverage")

    install(session, *packages, editable=True)
    session.run("pytest", "src", "tests/pytests", *session.posargs, env=env)


@nox.session
def integration(session: nox.Session) -> None:
    install(session, ".", "coverage[toml]", "fclogr", editable=True)
    packages_env = session.env.get("PACKAGES")
    packages = shlex.split(packages_env) if packages_env else INTEGRATION_PACKAGES
    with (
        coverage_run(session) as cov_env,
        # askalono does not like that the temporary directory is in the
        # gitignore'd .nox/ directory
        TemporaryDirectory(dir=os.environ.get("TMPDIR")) as tmp,
    ):
        cov_env["TMPDIR"] = str(tmp)
        for package in packages:
            with session.chdir(Path("tests/integration", package)):
                for script in ("integration-archive", "verify-license"):
                    session.run(
                        "bash",
                        "-x",
                        f"../../../contrib/{script}.sh",
                        env=cov_env,
                        external=True,
                    )


@nox.session(name="integration-test-build")
def integration_test_build(session: nox.Session):
    install(session, ".", "coverage[toml]", editable=True)
    with coverage_run(session) as cov_env:
        coverage_command = shlex.join(COVERAGE_COMMANDS.go_vendor_license)
        assert coverage_command[0]
        rpm_eval = Path("rpmeval.sh").resolve()
        with session.chdir("tests/integration/fzf"):
            # fmt: off
            session.run(
                "bash",
                "-x",
                str(rpm_eval),
                "-D", f"__go_vendor_license {coverage_command}",
                "-D", f"_specdir {Path.cwd()}",
                "-D", f"_sourcedir {Path.cwd()}",
                "--nodeps",
                "-ba", "fzf.spec",
                env=cov_env|{"RPM": "rpmbuild"},
            )
            # fmt: on


@nox.session
def coverage(session: nox.Session):
    install(session, "coverage[toml]")
    session.run("coverage", "combine", "--keep", *iglob(".nox/*/tmp/.coverage"))
    session.run("coverage", "xml")
    session.run("coverage", "html")
    session.run("coverage", "report", "--fail-under", COVERAGE_FAIL_UNDER)


@nox.session()
def covtest(session: nox.Session):
    session.run("rm", "-f", *iglob(".nox/*/tmp/.coverage*"), external=True)
    test_sessions = (f"test-{v}" for v in test.python)  # type: ignore[attr-defined]
    for target in test_sessions:
        session.notify(target, ["--cov"])


@nox.session(name="all")
def all_(session: nox.Session):
    lint(session)
    covtest(session)
    session.notify("integration")
    session.notify("integration-test-build")
    session.notify("coverage")


@nox.session(venv_backend="none")
def lint(session: nox.Session):
    """
    Run formatters, codeqa, and typing sessions
    """
    for notify in LINT_SESSIONS:
        session.notify(notify)


@nox.session
def codeqa(session: nox.Session):
    install(session, ".[codeqa]")
    session.run("ruff", "check", *session.posargs, *LINT_FILES)
    session.run("shellcheck", *iglob("contrib/*.sh"))
    session.run(
        "pymarkdownlnt", "scan", *iglob("*.md"), *iglob("docs/**/*.md", recursive=True)
    )
    session.run("reuse", "lint")


@nox.session
def formatters(session: nox.Session):
    install(session, ".[formatters]")
    posargs = session.posargs
    if IN_CI:
        posargs.append("--check")
    session.run("black", *posargs, *LINT_FILES)
    session.run("isort", *posargs, *LINT_FILES)


@nox.session
def typing(session: nox.Session):
    install(session, ".[typing]", editable=True)
    session.run("mypy", *LINT_FILES)


@nox.session
def bump(session: nox.Session):
    version = session.posargs[0]

    install(session, "build", "releaserr >= 0.1.dev123", "fclogr")
    session.run("releaserr", "--version")

    session.run("releaserr", "check-tag", version)
    session.run("releaserr", "ensure-clean")
    session.run("releaserr", "set-version", "-s", "file", version)

    install(session, ".")

    # Bump specfile
    # fmt: off
    session.run(
        "fclogr", "bump",
        "--new", version,
        "--comment", f"Release {version}.",
        SPECFILE,
    )
    # fmt: on

    # Bump changelog, commit, and tag
    git(session, "add", SPECFILE, f"src/{PROJECT}/__init__.py")
    session.run("releaserr", "clog", version, "--tag")
    # FIXME(anyone): Temporarily disable twine check.
    # It doesn't understand hatch 2.3 metadata.
    session.run(
        "releaserr", "build", "--backend", "generic", "--isolated", "--no-twine-check"
    )


@nox.session
def publish(session: nox.Session):
    # Setup
    install(session, "twine", "releaserr")
    session.run("releaserr", "--version")

    session.run("releaserr", "ensure-clean")

    # Upload to PyPI
    session.run("releaserr", "upload")

    # Post-release bump
    session.run("releaserr", "post-version", "-s", "file")
    git(session, "add", f"src/{PROJECT}/__init__.py")
    git(session, "commit", "-S", "-m", "Post-release version bump")


@nox.session
def srpm(session: nox.Session, posargs=None):
    install(session, "fclogr")
    posargs = posargs or session.posargs
    session.run("fclogr", "--debug", "dev-srpm", *posargs, SPECFILE)


@nox.session
def mockbuild(session: nox.Session):
    tmp = Path(session.create_tmp())
    srpm(session, ("-o", tmp, "--keep"))
    margs = [
        "mock",
        "--spec",
        str(Path(tmp, SPECFILE)),
        "--source",
        str(tmp),
        *session.posargs,
    ]
    if not session.interactive:
        margs.append("--verbose")
    session.run(*margs, external=True)


@nox.session
def releaserr(session: nox.Session):
    session.install("releaserr")
    session.run("releaserr", *session.posargs)


@nox.session
def mkdocs(session: nox.Session) -> None:
    install(session, "-r", "docs-requirements.txt")
    session.run("mkdocs", *(session.posargs if session.posargs else ["build"]))
